from typing import Annotated, Any

import gymnasium as gym
import numpy as np
import numpy.typing as npt
from gymnasium import Env, spaces

from .lssdp import LSSDPInstance, LSSDPSolution
from .network import SGNetwork  # type: ignore

ActType = int
ObsType = dict[str, Any]


class RidesEnv(Env):
    r"""The main Gymnasium class for implementing Reinforcement Learning Agents environments.

    The class encapsulates an environment with arbitrary behind-the-scenes dynamics through the :meth:`step` and :meth:`reset` functions.
    An environment can be partially or fully observed by single agents. For multi-agent environments, see PettingZoo.

    The main API methods that users of this class need to know are:

    - :meth:`step` - Updates an environment with actions returning the next agent observation, the reward for taking that actions,
      if the environment has terminated or truncated due to the latest action and information from the environment about the step, i.e. metrics, debug info.
    - :meth:`reset` - Resets the environment to an initial state, required before calling step.
      Returns the first agent observation for an episode and information, i.e. metrics, debug info.
    - :meth:`render` - Renders the environments to help visualise what the agent see, examples modes are "human", "rgb_array", "ansi" for text.
    - :meth:`close` - Closes the environment, important when external software is used, i.e. pygame for rendering, databases

    Environments have additional attributes for users to understand the implementation

    - :attr:`action_space` - The Space object corresponding to valid actions, all valid actions should be contained within the space.
    - :attr:`observation_space` - The Space object corresponding to valid observations, all valid observations should be contained within the space.
    - :attr:`spec` - An environment spec that contains the information used to initialize the environment from :meth:`gymnasium.make`
    - :attr:`metadata` - The metadata of the environment, i.e. render modes, render fps
    - :attr:`np_random` - The random number generator for the environment. This is automatically assigned during
      ``super().reset(seed=seed)`` and when assessing :attr:`np_random`.

    .. seealso:: For modifying or extending environments use the :py:class:`gymnasium.Wrapper` class

    Note:
        To get reproducible sampling of actions, a seed can be set with ``env.action_space.seed(123)``.
    """

    # # Set these in ALL subclasses
    # action_space: spaces.Space[ActType]
    # observation_space: spaces.Space[ObsType]

    # Created
    _np_random: np.random.Generator | None = None

    metadata = {"render_modes": ["human"], "render_fps": 1}

    def __init__(
        self,
        nstops: int | list[int] = [30, 45],
        nbuses: int | list[int] = [5, 20],
        nbuses_full_min: int = 1,
        truncate: bool = True,
        capacity: int | None = 90,
        speed: Annotated[float, "kmh"] = 17.0,
        dwell_time: Annotated[float, "min"] = 0.5,
        allow_retrospect: bool = False,
        demand_npeaks_max: int = 50,
        demand_peak_conc: float = 2.0,
        demnad_peak_size: float = 150.0,
        max_od_demand: float = 0.9,
        max_iters: int = 10000,
        render_mode: str | None = None,
    ):
        self._nstops = nstops
        self._nbuses = nbuses
        self._nbuses_full_min = nbuses_full_min
        self._truncate = truncate
        self._speed = speed
        self._dwell_time = dwell_time
        self._capacity = capacity
        self._allow_retrospect = allow_retrospect

        self._demand_npeaks_max = demand_npeaks_max
        self._demand_peak_conc = demand_peak_conc
        self._demand_peak_size = demnad_peak_size
        self._max_od_demand = max_od_demand

        self._max_iters = max_iters

        self._network = SGNetwork()

        _mat = lambda high: spaces.Box(
            0.0, high, shape=(self._nstops_max, self._nstops_max)
        )
        _oh = lambda size: spaces.MultiDiscrete([2] * size)

        self.observation_space = spaces.Dict(
            {
                "od_demand": _mat(10.0),
                "stops_exp": _oh(self._nstops_max),
                "nbuses_exp": spaces.Discrete(
                    self._nbuses_max - self._nbuses_full_min,
                    start=self._nbuses_full_min,
                ),
                "link_travel_time": _mat(120.0),
                "od_initial_travel_time": _mat(120.0),
                "od_travel_time": _mat(240.0),
                "od_relative_travel_time": _mat(10.0),
                "action_mask": _oh(self._nactions),
            }
        )

        self.action_space = spaces.Discrete(self._nactions)

        assert render_mode is None or render_mode in self.metadata["render_modes"]
        self.render_mode = render_mode

        self._inst: LSSDPInstance
        self._sol: LSSDPSolution

    def step(
        self, action: ActType
    ) -> tuple[ObsType, float, bool, bool, dict[str, Any]]:
        """Run one timestep of the environment's dynamics using the agent actions.

        When the end of an episode is reached (``terminated or truncated``), it is necessary to call :meth:`reset` to
        reset this environment's state for the next episode.

        .. versionchanged:: 0.26

            The Step API was changed removing ``done`` in favor of ``terminated`` and ``truncated`` to make it clearer
            to users when the environment had terminated or truncated which is critical for reinforcement learning
            bootstrapping algorithms.

        Args:
            action (ActType): an action provided by the agent to update the environment state.

        Returns:
            observation (ObsType): An element of the environment's :attr:`observation_space` as the next observation due to the agent actions.
                An example is a numpy array containing the positions and velocities of the pole in CartPole.
            reward (SupportsFloat): The reward as a result of taking the action.
            terminated (bool): Whether the agent reaches the terminal state (as defined under the MDP of the task)
                which can be positive or negative. An example is reaching the goal state or moving into the lava from
                the Sutton and Barton, Gridworld. If true, the user needs to call :meth:`reset`.
            truncated (bool): Whether the truncation condition outside the scope of the MDP is satisfied.
                Typically, this is a timelimit, but could also be used to indicate an agent physically going out of bounds.
                Can be used to end the episode prematurely before a terminal state is reached.
                If true, the user needs to call :meth:`reset`.
            info (dict): Contains auxiliary diagnostic information (helpful for debugging, learning, and logging).
                This might, for instance, contain: metrics that describe the agent's performance state, variables that are
                hidden from observations, or individual reward terms that are combined to produce the total reward.
                In OpenAI Gym <v26, it contains "TimeLimit.truncated" to distinguish truncation and termination,
                however this is deprecated in favour of returning terminated and truncated variables.
            done (bool): (Deprecated) A boolean value for if the episode has ended, in which case further :meth:`step` calls will
                return undefined results. This was removed in OpenAI Gym v26 in favor of terminated and truncated attributes.
                A done signal may be emitted for different reasons: Maybe the task underlying the environment was solved successfully,
                a certain timelimit was exceeded, or the physics simulation has entered an invalid state.
        """
        terminated = self._execute(action)

        # if self._sol._lss.is_valid():
        #     print(
        #         [
        #             1.0
        #             / self._inst.ass_trip_time
        #             * (self._inst.nbuses - self._sol._lss.nbuses),
        #             1.0
        #             / self._inst.trip_time(self._sol._lss.stops)
        #             * self._sol._lss.nbuses,
        #         ]
        #     )
        #     print(self._inst.ass_trip_time, self._inst.nbuses - self._sol._lss.nbuses)
        #     print(self._inst.ass_stops.stops)
        #     print(self._inst.trip_time(self._sol._lss.stops), self._sol._lss.nbuses)
        #     print(self._sol._lss.stops.stops)

        reward = self._sol._prev_obj - self._sol._obj

        if self.render_mode == "human":
            alignment = "".join(
                [
                    ("●" if selected else "○") + (" " if i % 5 == 0 and i != 0 else "")
                    for i, selected in enumerate(self._sol._lss.stops_binary)
                ]
            )

            if reward == 0:
                creward = f"{reward:9.4f}"
            elif reward > 0:
                creward = f"\033[32m{reward:9.4f}\033[0m"
            else:
                creward = f"\033[31m{reward:9.4f}\033[0m"

            stats = self._sol.stats
            _, ttd_mean, ttd_max, _ = stats["ttd"]
            _, lf_mean, lf_max, _ = stats["lf"]
            per_flow_exp = stats["per_flow_exp"]

            print(
                f" {self._step:2d} "
                f" {self._sol._obj:9.4f} "
                f" {creward} "
                f" {lf_mean:8.4f} / {lf_max:8.4f} "
                f" {ttd_mean:8.4f} / {ttd_max:8.4f} "
                f" {per_flow_exp: 6.4f} "
                f" {self._sol._lss.nbuses:3d} "
                f" {alignment}"
            )

            if terminated:
                print("-" * 137)

        self._step += 1

        return self._observation, reward, terminated, False, self._info

    def reset(
        self,
        *,
        seed: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> tuple[ObsType, dict[str, Any]]:
        """Resets the environment to an initial internal state, returning an initial observation and info.

        This method generates a new starting state often with some randomness to ensure that the agent explores the
        state space and learns a generalised policy about the environment. This randomness can be controlled
        with the ``seed`` parameter otherwise if the environment already has a random number generator and
        :meth:`reset` is called with ``seed=None``, the RNG is not reset.

        Therefore, :meth:`reset` should (in the typical use case) be called with a seed right after initialization and then never again.

        For Custom environments, the first line of :meth:`reset` should be ``super().reset(seed=seed)`` which implements
        the seeding correctly.

        .. versionchanged:: v0.25

            The ``return_info`` parameter was removed and now info is expected to be returned.

        Args:
            seed (optional int): The seed that is used to initialize the environment's PRNG (`np_random`).
                If the environment does not already have a PRNG and ``seed=None`` (the default option) is passed,
                a seed will be chosen from some source of entropy (e.g. timestamp or /dev/urandom).
                However, if the environment already has a PRNG and ``seed=None`` is passed, the PRNG will *not* be reset.
                If you pass an integer, the PRNG will be reset even if it already exists.
                Usually, you want to pass an integer *right after the environment has been initialized and then never again*.
                Please refer to the minimal example above to see this paradigm in action.
            options (optional dict): Additional information to specify how the environment is reset (optional,
                depending on the specific environment)

        Returns:
            observation (ObsType): Observation of the initial state. This will be an element of :attr:`observation_space`
                (typically a numpy array) and is analogous to the observation returned by :meth:`step`.
            info (dictionary):  This dictionary contains auxiliary information complementing ``observation``. It should be analogous to
                the ``info`` returned by :meth:`step`.
        """
        super().reset(seed=seed)

        if (
            options is None
            or "fix_problem" not in options
            or not options["fix_problem"]
        ):
            self._make_instance()

        self._sol = LSSDPSolution(self._inst)
        self._step = 1

        if self.render_mode == "human":
            spec_id = self.spec.id if self.spec is not None else ""

            print(f"  Env       : {spec_id}\n")
            self._inst.print_summary()

            div = "-" * 137
            print(
                f"{div}\n  #  Objective     Reward                   "
                f"LF                  TTD    % Exp  🚌  Alignment\n{div}"
            )

        return self._observation, self._info

    def render(self):  # type: ignore
        """Compute the render frames as specified by :attr:`render_mode` during the initialization of the environment.

        The environment's :attr:`metadata` render modes (`env.metadata["render_modes"]`) should contain the possible
        ways to implement the render modes. In addition, list versions for most render modes is achieved through
        `gymnasium.make` which automatically applies a wrapper to collect rendered frames.

        Note:
            As the :attr:`render_mode` is known during ``__init__``, the objects used to render the environment state
            should be initialised in ``__init__``.

        By convention, if the :attr:`render_mode` is:

        - None (default): no render is computed.
        - "human": The environment is continuously rendered in the current display or terminal, usually for human consumption.
          This rendering should occur during :meth:`step` and :meth:`render` doesn't need to be called. Returns ``None``.
        - "rgb_array": Return a single frame representing the current state of the environment.
          A frame is a ``np.ndarray`` with shape ``(x, y, 3)`` representing RGB values for an x-by-y pixel image.
        - "ansi": Return a strings (``str``) or ``StringIO.StringIO`` containing a terminal-style text representation
          for each time step. The text can include newlines and ANSI escape sequences (e.g. for colors).
        - "rgb_array_list" and "ansi_list": List based version of render modes are possible (except Human) through the
          wrapper, :py:class:`gymnasium.wrappers.RenderCollection` that is automatically applied during ``gymnasium.make(..., render_mode="rgb_array_list")``.
          The frames collected are popped after :meth:`render` is called or :meth:`reset`.

        Note:
            Make sure that your class's :attr:`metadata` ``"render_modes"`` key includes the list of supported modes.

        .. versionchanged:: 0.25.0

            The render function was changed to no longer accept parameters, rather these parameters should be specified
            in the environment initialised, i.e., ``gymnasium.make("CartPole-v1", render_mode="human")``
        """
        if self.render_mode is None:
            assert self.spec is not None
            gym.logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. gym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return None

        return None

    ##### User defined
    @property
    def _observation(self) -> ObsType:
        pad = self._pad

        return {
            "od_demand": pad(self._inst.demand),
            "stops_exp": pad(np.asarray(self._sol._lss.stops_binary)),
            "nbuses_exp": self._sol._lss.nbuses,
            "link_travel_time": pad(self._inst.travel_time),
            "od_initial_travel_time": pad(self._inst.base_ttd),
            "od_travel_time": pad(self._sol._ttd),
            "od_relative_travel_time": pad(self._sol._rel_ttd),
            "action_mask": self._action_mask,
        }

    @property
    def _action_mask(self):
        mask = np.ones(self._nactions, dtype=np.bool_)

        mask[1 : 1 + self._inst.nstops] = ~np.array(self._sol._lss.stops_binary)
        mask[-1] = self._sol._lss.nbuses < self._inst.nbuses - self._nbuses_full_min

        return mask

    @property
    def _info(self) -> dict[str, Any]:
        return {"nstops": self._inst.nstops, "nbuses": self._inst.nbuses}

    @property
    def _nstops_max(self) -> int:
        if isinstance(self._nstops, int):
            return self._nstops

        return self._nstops[1]

    @property
    def _nbuses_max(self) -> int:
        if isinstance(self._nbuses, int):
            return self._nbuses

        return self._nbuses[1]

    @property
    def _add_bus_action(self) -> int:
        return self._nactions - 1

    @property
    def _terminate_action(self) -> int:
        return 0

    @property
    def _nactions(self) -> int:
        return self._nstops_max + 2

    def _pad(self, mat: npt.NDArray[Any]) -> npt.NDArray[Any]:
        if mat.shape == (self._nstops_max,) * mat.ndim:
            return mat

        pad_width = self._nstops_max - mat.shape[0]
        return np.pad(mat, ((0, pad_width),) * mat.ndim)

    def _make_instance(self):
        self._inst = LSSDPInstance.from_network(
            self._network,
            self._nstops,
            self._nbuses,
            self._speed,
            self._dwell_time,
            self._demand_npeaks_max,
            self._demand_peak_conc,
            self._demand_peak_size,
            self._capacity,
            self._max_od_demand,
            self._truncate,
            self._max_iters,
            self.np_random,
        )

        self._sol = LSSDPSolution(self._inst)

    def _execute(self, action: int) -> bool:
        if not self.action_space.contains(action):
            raise ValueError("Invalid action")

        if action == self._terminate_action:
            self._sol.terminate()
            return True

        if action == self._add_bus_action:
            if self._sol._lss.nbuses >= self._inst.nbuses - self._nbuses_full_min:
                raise ValueError("Allocation to limited stop service exceeded limit")

            self._sol.add_bus()
            return False

        if action - 1 > self._inst.nstops:
            raise ValueError("Invalid action")

        if self._sol._lss.is_serving(action - 1):
            raise ValueError("Stop already served by limited stop service")

        if not self._allow_retrospect:
            if (
                not self._sol._lss.not_serving_any_stops()
                and action - 1 <= self._sol._lss.last_stop
            ):
                raise ValueError("Retrospective adding of stops not allowed")

        self._sol.toggle(action - 1)
        return False

    @property
    def sol(self) -> LSSDPSolution:
        return self._sol

    @property
    def inst(self) -> LSSDPInstance:
        return self._inst



import hashlib
from dataclasses import dataclass
from typing import Annotated, Any

import numpy as np
import numpy.typing as npt
from scipy.stats import multivariate_normal  # type: ignore
from tram import mat_linear_assign, mat_linear_congested_assign

from .entities import Service, StopSequence
from .network import SGNetwork


def calculate_stats(
    arr: npt.NDArray, sum: bool = True
) -> tuple[np.floating, np.floating, np.floating, np.floating]:
    values = (
        arr[np.triu_indices_from(arr, k=1)]
        if arr.ndim == 2 and arr.shape[0] == arr.shape[1]
        else arr
    )

    return (
        np.min(values),
        np.mean(values),
        np.max(values),
        np.sum(values) if sum else float("nan"),
    )


class LSSDPInstance:
    def __init__(
        self,
        travel_time: npt.NDArray[np.floating],
        demand: npt.NDArray[np.floating],
        nbuses: int,
        capacity: float,
        base_ttd: npt.NDArray[np.floating] = np.array([]),
        base_flow: npt.NDArray[np.floating] = np.array([]),
        base_obj: float = 0.0,
        max_iters: int = 10000,
        name: str | None = None,
    ):
        self.travel_time = travel_time
        self.demand = demand
        self.nbuses = nbuses
        self.capacity = capacity
        self.base_ttd = base_ttd
        self.base_flow = base_flow
        self.base_obj = base_obj
        self.max_iters = max_iters
        self.name = name

        self._id: str = ""

    def visualise(self) -> None:
        import math

        import matplotlib as mpl
        import matplotlib.pyplot as plt
        from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable

        def nanlower(arr):
            arr[np.tril_indices(arr.shape[0])] = np.nan
            return arr

        plt.rcParams.update(
            {
                "image.cmap": "RdYlBu",
                "axes.grid": True,
                "axes.grid.which": "major",
                "grid.color": "white",
                "grid.linewidth": 2,
                "axes.spines.left": False,
                "axes.spines.bottom": False,
                "axes.spines.top": False,
                "axes.spines.right": False,
                "xtick.labeltop": False,
                "xtick.labelbottom": False,
                "ytick.labelleft": False,
                "ytick.labelright": False,
                "xtick.top": False,
                "xtick.bottom": False,
                "ytick.left": False,
                "ytick.right": False,
            }
        )

        fig = plt.figure(figsize=(20, 10))
        gs = mpl.gridspec.GridSpec(6, 12, wspace=0.25, hspace=0.25)

        arr = fig.add_subplot(gs[0, :4])
        dep = fig.add_subplot(gs[1:5, 4])
        dem = fig.add_subplot(gs[1:5, :4])
        tt = fig.add_subplot(gs[1:5, 6:10])

        demand = self.demand
        demand_dep = demand.sum(axis=1)
        demand_arr = demand.sum(axis=0)
        demand_max = max(demand_dep.max(), demand_arr.max())
        interval = 5

        nstops = demand.shape[0]
        n_intervals = math.ceil(nstops / interval)

        dep.plot(demand_dep, range(nstops - 1, -1, -1))
        dep.set_ylim(-0.5, nstops - 0.5)
        dep.set_xlim(0, demand_max)

        arr.plot(demand_arr)
        arr.set_xlim(-0.5, nstops - 0.5)
        arr.set_ylim(0, demand_max)

        dem.imshow(nanlower(demand.copy()), cmap="OrRd", vmin=0)
        _ = dem.set_title("Demand", y=-0.1)
        _ = dem.set_xticks(np.arange(-0.5, n_intervals * interval - 0.5, interval))
        _ = dem.set_yticks(np.arange(-0.5, n_intervals * interval - 0.5, interval))
        for i in range(1, nstops + 1, interval):
            # _ = dem.plot([i-1, i-2], [i-1, i], 'k', lw=0.25)
            _ = dem.text(
                i - 2,
                i,
                i,
                rotation=-45,
                horizontalalignment="center",
                verticalalignment="center",
            )

        tt.imshow(nanlower(self.travel_time.copy()), cmap="OrRd", vmin=0)
        _ = tt.set_title("Travel time", y=-0.1)
        _ = tt.set_xticks(np.arange(-0.5, n_intervals * interval - 0.5, interval))
        _ = tt.set_yticks(np.arange(-0.5, n_intervals * interval - 0.5, interval))
        for i in range(1, nstops + 1, interval):
            # _ = tt.plot([i-1, i-2], [i-1, i], 'k', lw=0.25)
            _ = tt.text(
                i - 2,
                i,
                i,
                rotation=-45,
                horizontalalignment="center",
                verticalalignment="center",
            )

        plt.show()

    def calculate_id(self) -> None:
        h = hashlib.sha256()

        h.update(str(hash(tuple(self.travel_time.flatten().tolist()))).encode("utf8"))
        h.update(str(hash(tuple(self.demand.flatten().tolist()))).encode("utf8"))
        h.update(str(self.nbuses).encode("utf8"))
        h.update(str(self.capacity).encode("utf8"))

        self._id = h.hexdigest()[:10]

    def print_summary(self) -> None:
        print(
            f"  Instance  : {self._id}\n"
            f"  Name      : {self.name}\n"
            f"  Buses     : {self.nbuses}\n"
            f"  Stops     : {self.travel_time.shape[0]}\n"
            f"  Capacity  : {self.capacity}\n"
            f"  Objective : {self.base_obj:.4f}\n"
        )

        print("  \033[4mStats (min/avg/max/sum)\033[0m")
        print(
            "  OD Demand         : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f}\n"
            "  Dep Demand        : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f}\n"
            "  Arr Demand        : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f}\n"
            "  Link Travel time  : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f} (min)\n"
            "  Time to dest      : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f} (min)\n"
            "  Load factor       : {:7.2f} / {:7.2f} / {:7.2f} / {:8.2f} (%)\n".format(
                *calculate_stats(self.demand),
                *calculate_stats(self.demand.sum(axis=1)),
                *calculate_stats(self.demand.sum(axis=0)),
                *calculate_stats(self.travel_time, sum=False),
                *calculate_stats(self.base_ttd),
                *calculate_stats(
                    self.base_flow / self.nbuses / self.capacity * 100, sum=False
                ),
            )
        )

    @property
    def nstops(self) -> int:
        return self.travel_time.shape[0]

    # @cached_property
    @property
    def ass_stops(self) -> StopSequence:
        return StopSequence(list(range(self.nstops)))

    # @cached_property
    @property
    def ass_trip_time(self) -> float:
        return self.trip_time(self.ass_stops)

    # @cache
    def trip_time(self, stops: StopSequence) -> float:
        return sum(
            self.travel_time[from_][to_] for from_, to_ in zip(stops[:-1], stops[1:])
        )

    @staticmethod
    def from_network(
        network: SGNetwork,
        nstops: int | list[int],
        nbuses: int | list[int],
        speed: Annotated[float, "kmh"],
        dwell_time: Annotated[float, "min"],
        demand_npeaks_max: int,
        demand_peak_conc: float,
        demand_peak_size: float,
        capacity: float,
        max_od_demand: float,
        truncate: bool,
        max_iters: int,
        rng: np.random.Generator,
    ) -> "LSSDPInstance":
        # Generate fleet
        inst_nbuses = (
            nbuses
            if isinstance(nbuses, int)
            else rng.integers(nbuses[0], high=nbuses[1])
        )

        # Generate travel time matrix
        # _, distance, info = network.sample_test_route()
        if isinstance(nstops, int):
            _, distance, info = network.sample_real_route(
                nstops, truncate=truncate, rng=rng
            )
        elif isinstance(nstops, list):
            _, distance, info = network.sample_real_route(
                nstops[0], max_num_nodes=nstops[1], truncate=truncate, rng=rng
            )

        travel_time = np.triu(distance / speed / 1000 * 60 + dwell_time, 1)
        inst_nstops = travel_time.shape[0]

        # Generate demand matrix
        demand = rng.random(size=(inst_nstops, inst_nstops))

        if demand_npeaks_max > 1:
            for _ in range(rng.integers(1, high=demand_npeaks_max)):
                mean = sorted(rng.integers(1, high=inst_nstops + 1, size=(2,)))
                var = np.diag(
                    rng.random(size=(2,))
                    * inst_nstops
                    / rng.random()
                    / demand_peak_conc
                )

                x, y = np.mgrid[1 : inst_nstops + 1, 1 : inst_nstops + 1]
                pos = np.dstack((x, y))

                rv = multivariate_normal(mean, var, allow_singular=True)
                demand += rng.random() * rv.pdf(pos) * demand_peak_size

        demand = np.triu(demand, 1)
        demand /= np.max(demand)

        demand *= (
            rng.random()
            * capacity
            * max_od_demand
            / (inst_nstops + 1 - np.arange(inst_nstops).reshape(-1, 1))
        )  # Allow the demands for any OD pair to be up to 0.025x capacity

        inst = LSSDPInstance(
            travel_time=travel_time.astype(np.float32),
            demand=demand.astype(np.float32),
            nbuses=inst_nbuses,
            capacity=capacity,
        )

        # Calculate objective
        # out = mat_linear_assign(
        #     [inst.ass_stops.stops],
        #     [1 / inst.ass_trip_time * inst_nbuses],
        #     travel_time,
        #     demand,
        # )
        out = mat_linear_congested_assign(
            [inst.ass_stops.stops],
            [1 / inst.ass_trip_time * inst_nbuses],
            travel_time,
            demand,
            capacity,
            max_iters=max_iters,
        )

        inst.base_ttd = np.asarray(out[0], dtype=np.float32)
        inst.base_flow = np.asarray(out[1], dtype=np.float32)
        inst.base_obj = out[2]
        inst.name = info["name"]

        inst.calculate_id()

        return inst

    def __hash__(self) -> int:
        return self._id.__hash__()


class LSSDPSolution:
    def __init__(self, inst: LSSDPInstance) -> None:
        self._inst = inst

        self._lss = Service(nstops=inst.nstops, nbuses=1)
        self._prev_obj = 1.0
        self._obj = 1.0
        self._ttd = inst.base_ttd
        self._flow = inst.base_flow
        self._rel_ttd = np.triu(np.ones_like(self._ttd), 1)

    @property
    def _capacities(self) -> npt.NDArray[np.floating]:
        if not self._lss.is_valid():
            return np.array(
                [self._inst.nbuses * self._inst.capacity] * 3 * (self._inst.nstops - 1)
            )

        return np.array(
            (
                [(self._inst.nbuses - self._lss.nbuses) * self._inst.capacity]
                * 3
                * (self._inst.nstops - 1)
            )
            + (
                [self._lss.nbuses * self._inst.capacity]
                * 3
                * (len(self._lss.stops) - 1)
            )
        )

    @property
    def stats(
        self,
    ) -> dict[
        str, tuple[np.floating, np.floating, np.floating, np.floating] | np.floating
    ]:
        if not self._lss.is_valid():
            per_flow_exp = 0.0
        else:
            per_flow_exp = (
                self._flow[3 * (self._inst.nstops - 1) :].sum() / self._flow.sum()
            )

        return {
            "ttd": calculate_stats(self._ttd),
            "lf": calculate_stats(np.divide(self._flow, self._capacities)),
            "per_flow_exp": per_flow_exp,
        }

    def terminate(self) -> None:
        self._prev_obj = self._obj

    def toggle(self, stop: int) -> None:
        self._lss.toggle(stop)
        self._calculate_objective()

    def add_bus(self) -> None:
        self._lss.add_bus()
        self._calculate_objective()

    def _calculate_objective(self) -> None:
        if not self._lss.is_valid():
            self._prev_obj = self._obj
            self._obj = 1.0
            self._ttd = self._inst.base_ttd
            self._rel_ttd = np.triu(np.ones_like(self._ttd), 1)

            return

        alignments = [self._inst.ass_stops.stops, self._lss.stops.stops]
        frequencies = [
            1.0 / self._inst.ass_trip_time * (self._inst.nbuses - self._lss.nbuses),
            1.0 / self._inst.trip_time(self._lss.stops) * self._lss.nbuses,
        ]

        out = mat_linear_congested_assign(
            alignments,
            frequencies,
            self._inst.travel_time,
            self._inst.demand,
            self._inst.capacity,
            max_iters=self._inst.max_iters,
        )

        self._prev_obj = self._obj
        self._obj = out[2] / self._inst.base_obj
        self._ttd = np.asarray(out[0], dtype=np.float32)
        self._flow = np.asarray(out[1], dtype=np.float32)
        self._rel_ttd = np.divide(
            self._ttd,
            self._inst.base_ttd,
            out=np.zeros_like(self._inst.base_ttd),
            where=(self._inst.base_ttd != 0),
        )



import math
import os
import re
import urllib.request
from typing import List, Optional, Union

import geopandas as gpd
import numpy as np
import pandas as pd
import shapely

# def plot_route(stops, search_area, candidates):
#     wgs84 = pyproj.CRS("EPSG:4326")
#     utm = pyproj.CRS("EPSG:3414")

#     project = pyproj.Transformer.from_crs(utm, wgs84, always_xy=True).transform

#     stops = stops.to_crs("EPSG:4326")
#     #     candidates = candidates.to_crs('EPSG:4326')

#     lon, lat = stops.dissolve().centroid.iloc[0].xy
#     m = folium.Map(location=[lat[0], lon[0]], tiles="cartodbpositron", zoom_start=12)

#     folium.PolyLine(
#         list(zip(stops.geometry.y.tolist(), stops.geometry.x.tolist())), c="blue"
#     ).add_to(m)

#     for lat, lon in stops[["Latitude", "Longitude"]].values:
#         folium.CircleMarker(
#             location=[lat, lon],
#             radius=4,
#             stroke=True,
#             weight=2.5,
#             fill=True,
#             fill_color="#ffffff",
#             fill_opacity=1,
#         ).add_to(m)

#     x, y = transform(project, search_area).exterior.xy
#     folium.Polygon(list(zip(y, x)), fill=True, stroke=False, fill_opacity=0.5).add_to(m)

#     for lat, lon in candidates[["Latitude", "Longitude"]].values:
#         folium.CircleMarker(
#             location=[lat, lon],
#             radius=3,
#             color="#000000",
#             stroke=True,
#             weight=1,
#             fill=True,
#             fill_color="#ff0000",
#             fill_opacity=1,
#         ).add_to(m)

#     display(m)


def idx_max(lst, skip=0):
    assert skip < len(lst)
    return list(sorted(enumerate(lst), key=lambda x: x[1]))[-1 - skip][0]


class SGNetwork:
    _STOPS_TO_IGNORE = ["46211", "46219", "46239", "47711"]
    _ROUTES_TO_IGNORE = [
        "160",
        "170",
        "170X",
        "950",
        "101",
        "102",
        "107M",
        "11",
        "110",
        "111",
        "112",
        "113",
        "115",
        "116",
        "119",
        "120",
        "121",
        "122",
        "123M",
        "125",
        "127",
        "134",
        "138",
        "140",
        "142",
        "143M",
        "15",
        "150",
        "158",
        "160",
        "160M",
        "162M",
        "171",
        "173",
        "177",
        "179",
        "179A",
        "180",
        "181",
        "181M",
        "182",
        "182M",
        "183",
        "184",
        "189",
        "191",
        "194",
        "195",
        "199",
        "20",
        "200",
        "201",
        "222",
        "225G",
        "225W",
        "228",
        "229",
        "23",
        "231",
        "232",
        "235",
        "238",
        "24",
        "240",
        "241",
        "242",
        "243G",
        "243W",
        "246",
        "247",
        "248",
        "248M",
        "249",
        "251",
        "252",
        "253",
        "254",
        "255",
        "257",
        "258",
        "261",
        "262",
        "265",
        "269",
        "27",
        "272",
        "273",
        "282",
        "284",
        "285",
        "29",
        "291",
        "292",
        "293",
        "298",
        "300",
        "301",
        "302",
        "307",
        "315",
        "317",
        "324",
        "325",
        "329",
        "333",
        "334",
        "335",
        "34",
        "35",
        "354",
        "358",
        "359",
        "35M",
        "36",
        "37",
        "371",
        "372",
        "374",
        "381",
        "382",
        "382G",
        "382W",
        "384",
        "386",
        "40",
        "400",
        "401",
        "403",
        "405",
        "41",
        "410G",
        "410W",
        "42",
        "43M",
        "47",
        "49",
        "502",
        "518",
        "53",
        "60",
        "62",
        "63",
        "63M",
        "64",
        "66",
        "68",
        "69",
        "70M",
        "71",
        "73",
        "79",
        "800",
        "801",
        "803",
        "804",
        "805",
        "806",
        "807",
        "81",
        "811",
        "811T",
        "812",
        "812T",
        "82",
        "825",
        "83",
        "84",
        "857",
        "858",
        "859",
        "859A",
        "859B",
        "860",
        "882",
        "883",
        "883M",
        "89",
        "90",
        "900",
        "900A",
        "901",
        "901M",
        "903",
        "903M",
        "904",
        "91",
        "911",
        "912",
        "913",
        "92",
        "920",
        "922",
        "925M",
        "927",
        "92M",
        "94",
        "941",
        "944",
        "945",
        "947",
        "95",
        "950",
        "96",
        "962",
        "964",
        "965",
        "966",
        "972",
        "972M",
        "973",
        "974",
        "975",
        "979",
        "98",
        "983",
        "98M",
        "990",
    ]
    _WGS84 = "EPSG:4326"
    _SVY21 = "EPSG:3414"
    _BUS_STOPS_CSV_FN = "singapore_stops_grouped.csv"
    _BUS_ROUTES_CSV_FN = "singapore_routes.csv"
    _TTMAT_FN = "ttmat.npy"

    _TERMINAL_KW = ["INT", "TER"]
    _TERMINAL_REGEX_PATTERN = (
        "^(?!OPP|BEF|AFT)(?!.*FERRY|.*CARGO|.*LARKIN|.*JB|.*MARINE).* (?:"
        + "|".join(_TERMINAL_KW)
        + ")$"
    )
    _TERMINAL_PATTERN = (
        "^((?!(FERRY|POLICE|CARGO|CALTEX|FIRE|POWER|MARINE|RAILWAY)).)* "
        "(STN|INT|TER)((\/| EXIT).*|$)"
    )
    _EXTRACT_PATTERN = "((?:(?<!(?:OPP|BEF|AFT))[^/])*(?:STN|TER|INT))"

    def __init__(self, path_to_data: str = ".rides_sgnetwork") -> None:
        self.path_to_data = path_to_data

        if not os.path.exists(self.path_to_data):
            os.mkdir(self.path_to_data)
            print(f"Downloading {self.path_to_data}{os.path.sep}singapore_routes.csv")
            urllib.request.urlretrieve(
                "https://storage.googleapis.com/rides-env/singapore_routes.csv",
                f"{self.path_to_data}{os.path.sep}singapore_routes.csv",
            )
            print(
                f"Downloading {self.path_to_data}{os.path.sep}singapore_stops_grouped.csv"
            )
            urllib.request.urlretrieve(
                "https://storage.googleapis.com/rides-env/singapore_stops_grouped.csv",
                f"{self.path_to_data}{os.path.sep}singapore_stops_grouped.csv",
            )
            print(f"Downloading {self.path_to_data}{os.path.sep}ttmat.npy")
            urllib.request.urlretrieve(
                "https://storage.googleapis.com/rides-env/ttmat.npy",
                f"{self.path_to_data}{os.path.sep}ttmat.npy",
            )

        self._load_bus_stops()
        self._load_bus_routes()
        self._load_ttmat()

        self._find_terminals()
        # self._group_bus_stops()

    def _load_bus_stops(self) -> None:
        df = pd.read_csv(os.path.join(self.path_to_data, SGNetwork._BUS_STOPS_CSV_FN))

        self._bus_stops = (
            gpd.GeoDataFrame(
                df,
                crs=SGNetwork._WGS84,
                geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            )
            .to_crs(SGNetwork._SVY21)[
                lambda x: ~x.BusStopCode.isin(SGNetwork._STOPS_TO_IGNORE)
            ]
            .set_index("BusStopCode")
        )

    def _load_bus_routes(self) -> None:
        self._bus_routes = pd.read_csv(
            os.path.join(self.path_to_data, SGNetwork._BUS_ROUTES_CSV_FN)
        )
        self._bus_routes = self._bus_routes[
            lambda x: ~x.ServiceNo.isin(SGNetwork._ROUTES_TO_IGNORE)
        ]
        self._services = (
            self._bus_routes[["ServiceNo", "Direction"]]
            .drop_duplicates()
            .values.tolist()
        )

    def _load_ttmat(self) -> None:
        with open(os.path.join(self.path_to_data, SGNetwork._TTMAT_FN), "rb") as f:
            self._ttmat = np.load(f)

    def _find_terminals(self) -> None:
        self._terminals = self._bus_stops[
            lambda x: x.Description.str.contains(SGNetwork._TERMINAL_REGEX_PATTERN)
        ]

    def _group_bus_stops(self) -> None:
        bus_stops = self._bus_stops

        bus_stops = bus_stops.assign(
            Parent=lambda x: x.Description.str.extract(SGNetwork._EXTRACT_PATTERN)
        )

        terminals = bus_stops.Description.str.match(SGNetwork._TERMINAL_PATTERN)
        bus_stops.loc[~terminals, "Parent"] = np.nan

        replace_bus_int_ter = lambda x: x.str.replace(" (STN|INT|TER)$", "", regex=True)

        for i, stop in list(bus_stops[lambda x: x.Parent.isna()].iterrows()):
            if not np.isnan(stop.Parent):
                continue

            matches = bus_stops[bus_stops.within(stop.geometry.buffer(100))]

            if sum(matches.Parent.isna()) == 0:
                continue

            if len(matches) == 1:
                bus_stops.loc[i, "Parent"] = (
                    re.sub(r"^(OPP|AFT|BEF) ", "", stop.Description)
                    + " ("
                    + stop.RoadName
                    + ")"
                )
            elif sum(~matches.Parent.isna()) != 0:
                if matches.Parent.nunique() == 1:
                    bus_stops.loc[matches.index, "Parent"] = matches[
                        lambda x: ~x.Parent.isna()
                    ].Parent.iloc[0]
                else:
                    desc = replace_bus_int_ter(matches.Parent)[lambda x: ~x.isna()]

                    if desc.nunique() != 1:
                        raise ValueError

                    parent_name = (
                        desc.unique()[0]
                        + " "
                        + "/".join(
                            matches.Parent.str.replace(f"{desc.unique()[0]} ", "")[
                                lambda x: ~x.isna()
                            ]
                            .unique()
                            .tolist()
                        )
                    )
                    bus_stops.loc[matches.index, "Parent"] = parent_name
            else:
                parent_name = " / ".join(
                    matches.Description.str.replace("^(OPP|AFT|BEF)", "", regex=True)
                    .str.strip()
                    .unique()
                    .tolist()
                )

                if matches.RoadName.nunique() == 1:
                    parent_name += " (" + matches.RoadName.iloc[0] + ")"

                if parent_name in bus_stops.Parent.tolist():
                    bus_stops.loc[lambda x: x.Parent == parent_name, "Parent"] = (
                        parent_name + " [A]"
                    )
                    parent_name += " [B]"

                bus_stops.loc[matches.index, "Parent"] = parent_name

        self._bus_stops = bus_stops

    def _cache(self, route, distances, parents, circles) -> None:
        self._cached_route = route
        self._cached_distances = distances
        self._cached_parents = parents
        self._cached_circles = circles

    def _sample_terminal(self, n=1, min_distance=None, rng=None) -> List[str]:
        if rng is None:
            sampled = self._terminals.sample(n=n).index.tolist()
        else:
            idx = rng.choice(range(len(self._terminals)), replace=False, size=2)
            sampled = self._terminals.iloc[idx].index.tolist()

        if min_distance is None:
            return sampled

        assert n == 2

        if self._get_distance(*sampled) < min_distance:
            return self._sample_terminal(n=n, min_distance=min_distance, rng=rng)

        return sampled

    def _get_distance(self, from_: Union[str, List[str]], to_: Union[str, List[str]]):
        if isinstance(from_, list):
            from_ = self._bus_stops.loc[from_]["index"].tolist()
        else:
            from_ = self._bus_stops.loc[from_]["index"]

        if isinstance(to_, list):
            to_ = self._bus_stops.loc[to_]["index"].tolist()
        else:
            to_ = self._bus_stops.loc[to_]["index"]

        return self._ttmat[from_, to_]

    def _calc_search_radius(
        self,
        distance: float,
        max_search_radius=float("inf"),
        search_distance_factor=0.2,
    ) -> float:
        if math.isinf(max_search_radius):
            return distance * search_distance_factor

    def _get_search_area(
        self,
        from_: str,
        to_: str,
        max_search_radius=float("inf"),
        search_distance_factor=0.2,
    ) -> shapely.geometry.Polygon:
        distance = self._get_distance(from_, to_)

        center = (
            self._bus_stops.loc[from_]
            .geometry.buffer(distance / 2)
            .buffer(1)
            .intersection(
                self._bus_stops.loc[to_].geometry.buffer(distance / 2).buffer(1)
            )
        )  # Buffer 1m for cases where two stops are on the same road, i.e. intersection = ø

        search_radius = self._calc_search_radius(
            distance,
            max_search_radius=max_search_radius,
            search_distance_factor=search_distance_factor,
        )
        search_area = center.buffer(search_radius)

        return search_area

    def _search_candidates(
        self,
        pos: int,
        from_: str,
        to_: str,
        plot: bool = True,
        max_path_detour=1.2,
        max_link_detour=1.3,
        max_search_radius=float("inf"),
        search_distance_factor=0.2,
    ) -> gpd.GeoDataFrame:
        search_area = self._get_search_area(
            from_,
            to_,
            max_search_radius=max_search_radius,
            search_distance_factor=search_distance_factor,
        )
        # for i, circle in enumerate(self._cached_circles):
        #     if i != pos:
        #         search_area = search_area - circle

        candidates = (
            self._bus_stops.loc[~self._bus_stops.index.isin(self._cached_route)][
                lambda x: x.within(search_area)
            ][lambda x: ~x["Parent"].isin(self._cached_parents)]
        ).assign(
            straight_line_dist1=lambda x: x.distance(
                self._bus_stops.loc[from_].geometry
            ),
            straight_line_dist2=lambda x: x.distance(self._bus_stops.loc[to_].geometry),
            actual_dist1=lambda x: self._get_distance(from_, x.index.tolist()),
            actual_dist2=lambda x: self._get_distance(x.index.tolist(), to_),
            detour1=lambda x: x.actual_dist1 / x.straight_line_dist1,
            detour2=lambda x: x.actual_dist2 / x.straight_line_dist2,
            detour=lambda x: (x.actual_dist1 + x.actual_dist2)
            / (x.straight_line_dist1 + x.straight_line_dist2),
            prop=lambda x: x.actual_dist1 / x.actual_dist2,
        )[
            lambda x: x.detour <= max_path_detour
        ][
            lambda x: x.detour1 <= max_link_detour
        ][
            lambda x: x.detour2 <= max_link_detour
        ]

        if plot:
            plot_route(self._bus_stops.loc[self._cached_route], search_area, candidates)

        return candidates

    def _select_and_commit(
        self,
        pos: int,
        from_: str,
        to_: str,
        candidates: gpd.GeoDataFrame,
        rng=None,
        sample_candidate_threshold_distance=5000,
    ) -> None:
        if self._cached_distances[pos] >= sample_candidate_threshold_distance:
            if rng is None:
                selected = candidates.sample().iloc[0]
            else:
                idx = rng.choice(range(len(candidates)))
                selected = candidates.iloc[idx]
        else:
            candidates = candidates.sort_values(by="detour")
            selected = candidates.iloc[0]

        self._cached_circles[pos] = self._get_circle(from_, selected.name)
        self._cached_circles.insert(pos + 1, self._get_circle(selected.name, to_))

        self._cached_distances[pos] = self._get_distance(from_, selected.name)
        self._cached_distances.insert(pos + 1, self._get_distance(selected.name, to_))

        self._cached_route.insert(pos + 1, selected.name)
        self._cached_parents.append(selected["Parent"])

        return self._cached_route

    def _get_circle(self, from_, to_):
        ctr = self._bus_stops.loc[[from_, to_]].dissolve().centroid.iloc[0]
        distance = self._get_distance(from_, to_)
        shape = ctr.buffer(distance / 2)

        return shape

    def _break_longest_link(
        self,
        route,
        rng=None,
        failed_attempts=0,
        sample_candidate_threshold_distance=5000,
        max_path_detour=1.2,
        max_link_detour=1.3,
        max_search_radius=float("inf"),
        search_distance_factor=0.2,
    ) -> List[str]:
        if "_cached_route" not in self.__dict__ or route != self._cached_route:
            distances = [
                self._get_distance(*link) for link in zip(route[:-1], route[1:])
            ]
            parents = self._bus_stops.loc[route].Parent.unique().tolist()
            circles = [self._get_circle(*route)]
            self._cache(route, distances, parents, circles)

        while True:
            # try:
            pos = idx_max(self._cached_distances, skip=failed_attempts)
            from_, to_ = self._cached_route[pos], self._cached_route[pos + 1]
            # except:
            #     raise RuntimeError

            candidates = self._search_candidates(
                pos,
                from_,
                to_,
                max_path_detour=max_path_detour,
                max_link_detour=max_link_detour,
                max_search_radius=max_search_radius,
                search_distance_factor=search_distance_factor,
            )

            if len(candidates) == 0:
                failed_attempts += 1
                continue

            return (
                self._select_and_commit(
                    pos,
                    from_,
                    to_,
                    candidates,
                    rng=rng,
                    sample_candidate_threshold_distance=sample_candidate_threshold_distance,
                ),
                failed_attempts,
            )

    def _get_ttmat(self, route):
        idx = self._bus_stops.loc[route]["index"].tolist()
        return np.triu(self._ttmat[idx][:, idx], 1)

    def generate_gmap_link(self, route):
        url = "https://www.google.com/maps/dir/"

        for lat, lon in self._bus_stops.loc[route][["Latitude", "Longitude"]].values:
            url += f"{lat},{lon}/"

        return url

    def generate_trunk_route(
        self,
        min_num_nodes: int,
        max_num_nodes: Optional[int] = None,
        rng=None,
        min_terminal_distance: float = 5000.0,
        sample_candidate_threshold_distance: float = 5000.0,
        max_path_detour: float = 1.2,
        max_link_detour: float = 1.3,
        min_search_radius: float = 0.0,
        max_search_radius: float = float("inf"),
        search_distance_factor: float = 0.2,
    ):
        if max_num_nodes is None:
            max_num_nodes = min_num_nodes

        if rng is not None:
            target_num_nodes = rng.integers(min_num_nodes, max_num_nodes, endpoint=True)
        else:
            target_num_nodes = np.random.randint(min_num_nodes, max_num_nodes + 1)

        route = self._sample_terminal(n=2, min_distance=min_terminal_distance, rng=rng)

        failed_attempts = 0
        while len(route) < target_num_nodes:
            # try:
            route, failed_attempts = self._break_longest_link(
                route,
                rng=rng,
                failed_attempts=failed_attempts,
                sample_candidate_threshold_distance=sample_candidate_threshold_distance,
                max_path_detour=max_path_detour,
                max_link_detour=max_link_detour,
                max_search_radius=max_search_radius,
                search_distance_factor=search_distance_factor,
            )
            # except:
            #     break

        if len(route) > min_num_nodes:
            return route, self._get_ttmat(route), {}

        print("Retrying...")
        return self.generate_trunk_route(min_num_nodes, max_num_nodes, rng=rng)

    def sample_test_route(self):
        should_truncate = False

        route = ["A", "B", "C"]
        ttmat = np.asarray([[0, 850, 850], [0, 0, 1700], [0, 0, 0]])
        service = "TEST"
        direction = "A"

        return (
            route,
            ttmat,
            {
                "service": service,
                "direction": direction,
                "name": f"{service}-D{direction}"
                + (" (Truncated)" if should_truncate else ""),
                "truncated": should_truncate,
            },
        )

    def sample_real_route(
        self,
        min_num_nodes: int,
        max_num_nodes: Optional[int] = None,
        truncate: bool = True,
        rng=None,
    ):
        cannot_truncate = not truncate

        if max_num_nodes is None:
            max_num_nodes = min_num_nodes

        if rng is not None:
            target_num_nodes = rng.integers(min_num_nodes, max_num_nodes, endpoint=True)
        else:
            target_num_nodes = np.random.randint(min_num_nodes, max_num_nodes + 1)

        if rng is None:
            idx = np.random.choice(range(len(self._services)))
        else:
            idx = rng.choice(range(len(self._services)))

        service, direction = self._services[idx]
        route = (
            self._bus_routes[lambda x: x.ServiceNo == service][
                lambda x: x.Direction == direction
            ]
            .sort_values(by="StopSequence")
            .BusStopCode.tolist()
        )

        should_truncate = len(route) > target_num_nodes

        if len(route) < min_num_nodes or (should_truncate and cannot_truncate):
            return self.sample_real_route(
                min_num_nodes, max_num_nodes=max_num_nodes, rng=rng, truncate=truncate
            )

        if should_truncate:
            if rng is None:
                idx = np.random.choice(
                    range(len(route)), target_num_nodes, replace=False
                )
            else:
                idx = rng.choice(range(len(route)), target_num_nodes, replace=False)

            route = [route[i] for i in sorted(idx)]

        return (
            route,
            self._get_ttmat(route),
            {
                "service": service,
                "direction": direction,
                "name": f"{service}-D{direction}"
                + (" (Truncated)" if should_truncate else ""),
                "truncated": should_truncate,
            },
        )



from dataclasses import dataclass


@dataclass
class StopSequence:
    stops: list[int]

    def __len__(self) -> int:
        return self.stops.__len__()

    def __hash__(self) -> int:
        return tuple(self.stops).__hash__()

    def __getitem__(self, index):
        return self.stops.__getitem__(index)


class Service:
    def __init__(self, nstops: int, nbuses: int):
        self.stops_binary = [False for stop in range(nstops)]
        self.nbuses = nbuses
        self.nstops = nstops

        self.last_stop = -1

    @property
    def stops(self) -> StopSequence:
        return StopSequence([i for i, served in enumerate(self.stops_binary) if served])

    def is_valid(self) -> bool:
        return sum(self.stops_binary) >= 2 and self.nbuses >= 1

    def is_serving(self, stop: int) -> bool:
        return self.stops_binary[stop]

    def not_serving_any_stops(self) -> bool:
        return sum(self.stops_binary) == 0

    def add_bus(self) -> None:
        self.nbuses += 1

    def toggle(self, stop: int) -> None:
        self.stops_binary[stop] = False if self.stops_binary[stop] else True

        if stop >= self.last_stop:
            if self.stops_binary[stop]:
                self.last_stop = stop
            elif sum(self.stops_binary) == 0:
                self.last_stop = -1
            else:
                for i in range(self.last_stop, -1, -1):
                    if self.stops_binary[i]:
                        self.last_stop = i
                        break
